//////////////////////////////////////////////////////////
//								
//	Simple Program to Illustrate Layered Rendering		
//								
//		    Suryakant Patidar 'skp'			
//		 	skp@research.iiit.ac.in		
//			http://cvit.iiit.ac.in	
//
/////////////////////////////////////////////////////////

//To enable all the Extensions to be used
#define	GL_GLEXT_PROTOTYPES

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>

#include <GL/glut.h>

GLuint fbo;		//Holds the layered Frame Buffer Object
GLuint colorTex;	//3D Texture == 2D Array EXT

GLuint	vertexShader, geometryShader, shader;		// Shader for first pass
GLuint	vertexShader2, fragmentShader2, shader2;	// Shader for second pass

int width = 512;
int height = 256;

void keys( unsigned char key, int x, int y) 
{
	switch(key) 
	{
		case 27:
			exit(0);
			break;
	};
	glutPostRedisplay();
}

void firstPass() 
{
	//Activate Frame buffer Object
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo);
	//Set the correct Viewport, as per your texture size
	glViewport( 0, 0, 256, 256);
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 40.0f, 1.0f, 0.5f, 15.0f);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef( 0.0f, 0.0f, -8.0f);

	//Enable Shader, which routes the triangle to 2 Layers with diff color and orientation
	glUseProgram(shader);
	
	glBegin(GL_TRIANGLES);
	glVertex3d(-2.0,-2.0,0.0);
	glVertex3d(2.0,0.0,0.0);
	glVertex3d(-2.0,2.0,0.0);
	glEnd();

	//Restore the default "System Frame Buffer"
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0);
}

void draw() 
{
	//To render to Layered FBO
	firstPass();

	//Restore the Viewport for Second pass
	glViewport( 0, 0, width, height);
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho( 0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
		
	//Binding a different shader which shows the 2 Layers we rendered to in firstPass
	glUseProgram( shader2 );
	//Bind the 3D Texture 
	glBindTexture( GL_TEXTURE_2D_ARRAY_EXT, colorTex);
	glColor3f( 1.0f, 1.0f, 0.0f);
		
	//Screen Size Quad Drawn
	glBegin( GL_QUADS);
	glVertex2f( 0.0f, 0.0f);
	glVertex2f( 1.0f, 0.0f);
	glVertex2f( 1.0f, 1.0f);
	glVertex2f( 0.0f, 1.0f);
	glEnd();

	glBindTexture( GL_TEXTURE_2D_ARRAY_EXT, 0);

	glutSwapBuffers();
}

void reshape( int w, int h) 
{
	width = w;
	height = h;

	glViewport( 0, 0, w, h);
}

char *textFileRead(char *fn) 
{
	FILE *fp;
	char *content = NULL;

	int f,count;
	f = open(fn, O_RDONLY);

	count = lseek(f, 0, SEEK_END);

	close(f);

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}

void printInfoLog(GLhandleARB obj)
{
	int infologLength = 0;
	int charsWritten = 0;
	char *infoLog;
	glGetObjectParameterivARB(obj, GL_OBJECT_INFO_LOG_LENGTH_ARB, &infologLength);
	if (infologLength > 0)
	{
		infoLog = (char *)malloc(infologLength);
		glGetInfoLogARB(obj, infologLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
		free(infoLog);
	}
}

void initShaders() 
{
	char *vs,*fs,*gs;

	vertexShader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	geometryShader = glCreateShaderObjectARB(GL_GEOMETRY_SHADER_EXT);

	vs = textFileRead("vs.glsl");
	gs = textFileRead("gs.glsl");

	const char * vv = vs;
	const char * gg = gs;

	glShaderSourceARB(vertexShader, 1, &vv,NULL);
	glShaderSourceARB(geometryShader, 1, &gg,NULL);

	//printf ("The VS Code :: \n%s\n\n",vv);
	//printf ("The GS Code :: \n%s\n\n",gg);
	free(vs);free(gs);

	glCompileShaderARB(vertexShader);
	//printInfoLog( vertexShader );
	glCompileShaderARB(geometryShader);
	//printInfoLog( geometryShader );

	shader = glCreateProgramObjectARB();
	glAttachObjectARB( shader, vertexShader);
	glAttachObjectARB( shader, geometryShader);

	glLinkProgramARB( shader );

	// Loading the 2nd Shader Set
	vertexShader2 = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	fragmentShader2 = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

	vs = textFileRead("vs2.glsl");
	fs = textFileRead("fs2.glsl");

	vv = vs;
	const char * ff = fs;

	glShaderSourceARB(vertexShader2, 1, &vv,NULL);
	glShaderSourceARB(fragmentShader2, 1, &ff,NULL);

	//printf ("The VS2 Code :: \n%s\n\n",vv);
	//printf ("The FS2 Code :: \n%s\n\n",ff);
	free(vs);free(fs);

	glCompileShaderARB(vertexShader2);
	//printInfoLog( vertexShader2 );
	glCompileShaderARB(fragmentShader2);
	//printInfoLog( fragmentShader2 );

	shader2 = glCreateProgramObjectARB();
	glAttachObjectARB( shader2, vertexShader2);
	glAttachObjectARB( shader2, fragmentShader2);

	glLinkProgramARB( shader2 );
}


void init() 
{
	//Generate an FBO and Texture ID to hold the 3D Texture
	glGenTextures( 1, &colorTex);
	glGenFramebuffersEXT( 1, &fbo);
	
	glBindTexture( GL_TEXTURE_2D_ARRAY_EXT, colorTex);
	//Create a 3D Texture with RGBA8 
	glTexImage3D( GL_TEXTURE_2D_ARRAY_EXT, 0, GL_RGBA8, 256, 256, 2, 0, GL_RGBA, GL_FLOAT, 0);
	glTexParameteri( GL_TEXTURE_2D_ARRAY_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri( GL_TEXTURE_2D_ARRAY_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri( GL_TEXTURE_2D_ARRAY_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri( GL_TEXTURE_2D_ARRAY_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	glBindTexture( GL_TEXTURE_2D_ARRAY_EXT, 0);

	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo);
	//Attach the 3D texture (which contains 4 Layers) to the FBO
	glFramebufferTextureEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, colorTex, 0);
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0);

	//Load and link 2 sets of shaders 
	initShaders();

	//Informing the Geometry Shader that there will be 6 vertices emitted per primitive
	glProgramParameteriEXT( shader, GL_GEOMETRY_VERTICES_OUT_EXT, 6);
	//Informing the Geometry Shader that the input type to GS will be Triangles
	glProgramParameteriEXT( shader, GL_GEOMETRY_INPUT_TYPE_EXT, GL_TRIANGLES);
	//Informing the Geometry Shader that the Output type to GS will be Triangle Strips
	glProgramParameteriEXT( shader, GL_GEOMETRY_OUTPUT_TYPE_EXT, GL_TRIANGLE_STRIP);

	GLint success = 0;
	glGetProgramiv( shader, GL_LINK_STATUS, &success);
	if (!success)
	{
		printf ("Shader Program not Linked Correctly\nExiting\n");
		exit(0);
	}

	//Enable Depth Testing
	glEnable(GL_DEPTH_TEST);
}

int main( int argc, char **argv) 
{
	glutInit( &argc, argv);
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize( width, height);
	glutCreateWindow( "Render Layers : Simple");

	glutDisplayFunc( draw);
	glutReshapeFunc( reshape);
	glutKeyboardFunc(keys);
	glutIdleFunc(draw);

	init();
	glutMainLoop();
	return 0;
}
